# -*- coding: utf-8 -*-

"""Core functionality

Author: Gertjan van den Burg

"""

import os
import warnings
import sqlite3
import shutil

from .exceptions import (
    DatabaseNotFound,
    DatabaseVersionNotFound,
)
from .models import (
    Attachment,
    MMSMessageRecord,
    Quote,
    Recipient,
    RecipientId,
    SMSMessageRecord,
    Thread,
)
from .html import dump_thread
from .html_colors import get_random_color


def check_backup(backup_dir):
    """Check that we have the necessary files"""
    if not os.path.join(backup_dir, "database.sqlite"):
        raise DatabaseNotFound
    if not os.path.join(backup_dir, "DatabaseVersion.sbf"):
        raise DatabaseVersionNotFound
    with open(os.path.join(backup_dir, "DatabaseVersion.sbf"), "r") as fp:
        version_str = fp.read()

    # We have only ever seen database version 23, so we don't proceed if it's
    # not that. Testing and pull requests welcome.
    version = version_str.split(":")[-1].strip()
    if not version in ["23", "65", "80"]:
        warnings.warn(
            f"Warning: Found untested Signal database version: {version}."
        )
    return version


def get_color(db, recipient_id):
    """ Extract recipient color from the database """
    query = db.execute(
        "SELECT color FROM recipient_preferences WHERE recipient_ids=?",
        (recipient_id,),
    )
    color = query.fetchone()[0]
    return color


def make_recipient_v23(db, recipient_id):
    """ Create a Recipient instance from a given recipient id (db version 23)"""
    if recipient_id.startswith("__textsecure_group__"):
        qry = db.execute(
            "SELECT title FROM groups WHERE group_id=?", (recipient_id,)
        )
        label = qry.fetchone()
        isgroup = True
    else:
        qry = db.execute(
            "SELECT system_display_name FROM recipient_preferences WHERE recipient_ids=?",
            (recipient_id,),
        )
        label = qry.fetchone()
        isgroup = False

    name = recipient_id if label[0] is None else label[0]
    color = get_color(db, recipient_id)

    rid = RecipientId(recipient_id)
    return Recipient(rid, name=name, color=color, isgroup=isgroup)


def make_recipient_v80(db, recipient_id):
    """Create a Recipient instance from a recipient id (db version 65, 80)"""
    qry = db.execute(
        "SELECT group_id, system_display_name, profile_joined_name, color from recipient where _id=?",
        (recipient_id,),
    )
    groupid, name, joined_name, color = qry.fetchone()
    if color is None:
        color = get_random_color()
    isgroup = groupid is not None
    if isgroup:
        qry = db.execute(
            "SELECT title FROM groups WHERE group_id=?", (groupid,)
        )
        name = qry.fetchone()[0]

    if name is None:
        if joined_name is None:
            name = str(recipient_id)
        name = joined_name

    rid = RecipientId(recipient_id)
    return Recipient(rid, name=name, color=color, isgroup=isgroup)


def make_recipient(db, recipient_id, version=None):
    if version == "23":
        return make_recipient_v23(db, recipient_id)
    elif version in ["65", "80"]:
        return make_recipient_v80(db, recipient_id)
    else:
        warnings.warn(
            f"Untested database version {version}, defaulting to latest known working version."
        )
        return make_recipient_v80(db, recipient_id)


def get_sms_records(db, thread, version=None):
    """ Collect all the SMS records for a given thread """
    sms_records = []
    sms_qry = db.execute(
        "SELECT _id, address, date, date_sent, body, type "
        "FROM sms WHERE thread_id=?",
        (thread._id,),
    )
    qry_res = sms_qry.fetchall()
    for _id, address, date, date_sent, body, _type in qry_res:
        sms = SMSMessageRecord(
            _id=_id,
            addressRecipient=make_recipient(db, address, version=version),
            recipient=thread.recipient,
            dateSent=date_sent,
            dateReceived=date,
            threadId=thread._id,
            body=body,
            _type=_type,
        )
        sms_records.append(sms)
    return sms_records


def get_attachment_filename(_id, unique_id, backup_dir, thread_dir):
    """ Get the absolute path of an attachment, warn if it doesn't exist"""
    fname = f"Attachment_{_id}_{unique_id}.bin"
    source = os.path.abspath(os.path.join(backup_dir, fname))
    if not os.path.exists(source):
        warnings.warn(
            f"Warning: couldn't find attachment {source}. Maybe it was deleted?"
        )
        return None

    # Copying here is a bit of a side-effect
    target_dir = os.path.abspath(os.path.join(thread_dir, "attachments"))
    os.makedirs(target_dir, exist_ok=True)
    target = os.path.join(target_dir, fname)
    shutil.copy(source, target)
    return target


def add_mms_attachments(db, mms, backup_dir, thread_dir):
    """ Add all attachment objects to MMS message """
    qry = db.execute(
        "SELECT _id, ct, unique_id, voice_note, width, height, quote "
        "FROM part WHERE mid=?",
        (mms._id,),
    )
    for _id, ct, unique_id, voice_note, width, height, quote in qry:
        a = Attachment(
            contentType=ct,
            unique_id=unique_id,
            fileName=get_attachment_filename(
                _id, unique_id, backup_dir, thread_dir
            ),
            voiceNote=voice_note,
            width=width,
            height=height,
            quote=quote,
        )
        mms.attachments.append(a)


def get_mms_records(db, thread, recipients, backup_dir, thread_dir, version=None):
    """ Collect all MMS records for a given thread """
    mms_records = []
    qry = db.execute(
        "SELECT _id, address, date, date_received, body, quote_id, "
        "quote_author, quote_body, msg_box FROM mms WHERE thread_id=?",
        (thread._id,),
    )
    qry_res = qry.fetchall()
    for (
        _id,
        address,
        date,
        date_received,
        body,
        quote_id,
        quote_author,
        quote_body,
        msg_box,
    ) in qry_res:
        quote = None
        if quote_id:
            quote_auth = next(
                (
                    r
                    for r in recipients
                    if str(r.recipientId._id) == str(quote_author)
                ),
                None,
            )
            if quote_auth is None:
                # Quote is from someone who isn't a recipient (e.g. a friend
                # quotes a third person in a group). We'll just create a
                # recipient object for this person.
                rid = RecipientId(quote_author)
                quote_auth = Recipient(
                    rid, quote_author, color=None, isgroup=False
                )
            quote = Quote(_id=quote_id, author=quote_auth, text=quote_body)

        mms = MMSMessageRecord(
            _id=_id,
            addressRecipient=make_recipient(db, address, version=version),
            recipient=thread.recipient,
            dateSent=date,
            dateReceived=date_received,
            threadId=thread._id,
            body=body,
            quote=quote,
            attachments=[],
            _type=msg_box,
        )
        mms_records.append(mms)

    for mms in mms_records:
        add_mms_attachments(db, mms, backup_dir, thread_dir)

    return mms_records


def populate_thread(db, thread, recipients, backup_dir, thread_dir, 
        version=None):
    """ Populate a thread with all corresponding messages """
    sms_records = get_sms_records(db, thread, version=version)
    mms_records = get_mms_records(
        db, thread, recipients, backup_dir, thread_dir, version=version
    )
    thread.sms = sms_records
    thread.mms = mms_records


def process_backup(backup_dir, output_dir):
    """ Main functionality to convert database into HTML """

    # Verify backup and open database
    db_version = check_backup(backup_dir)
    db_file = os.path.join(backup_dir, "database.sqlite")
    db_conn = sqlite3.connect(db_file)
    db = db_conn.cursor()

    # Start by getting the Threads from the database
    query = db.execute("SELECT _id, recipient_ids FROM thread")
    threads = query.fetchall()

    # Now turn the recipients from the threads into Recipient classes
    recipients = []
    for _, recipient_id in threads:
        r = make_recipient(db, recipient_id, version=db_version)
        recipients.append(r)

    # Combine the recipient objects and the thread info into Thread objects
    for (_id, _), recipient in zip(threads, recipients):
        t = Thread(_id=_id, recipient=recipient)
        thread_dir = os.path.join(output_dir, t.name.replace(" ", "_"))
        populate_thread(db, t, recipients, backup_dir, thread_dir, version=db_version)
        dump_thread(t, thread_dir)

    db.close()



# -*- coding: utf-8 -*-

"""Colors extracted from Signal source

"""

import random

CRIMSON = "#CC163D"
CRIMSON_TINT = "#EDA6AE"
CRIMSON_SHADE = "#8A0F29"
VERMILLION = "#C73800"
VERMILLION_TINT = "#EBA78E"
VERMILLION_SHADE = "#872600"
BURLAP = "#746C53"
BURLAP_TINT = "#C4B997"
BURLAP_SHADE = "#58513C"
FOREST = "#3B7845"
FOREST_TINT = "#8FCC9A"
FOREST_SHADE = "#2B5934"
WINTERGREEN = "#1C8260"
WINTERGREEN_TINT = "#9BCFBD"
WINTERGREEN_SHADE = "#36544A"
TEAL = "#067589"
TEAL_TINT = "#A5CAD5"
TEAL_SHADE = "#055968"
BLUE = "#336BA3"
BLUE_TINT = "#ADC8E1"
BLUE_SHADE = "#285480"
INDIGO = "#5951C8"
INDIGO_TINT = "#C2C1E7"
INDIGO_SHADE = "#4840A0"
VIOLET = "#862CAF"
VIOLET_TINT = "#CDADDC"
VIOLET_SHADE = "#6B248A"
PLUMB = "#A23474"
PLUMB_TINT = "#DCB2CA"
PLUMB_SHADE = "#881B5B"
TAUPE = "#895D66"
TAUPE_TINT = "#CFB5BB"
TAUPE_SHADE = "#6A4E54"
STEEL = "#6B6B78"
STEEL_TINT = "#BEBEC6"
STEEL_SHADE = "#5A5A63"
ULTRAMARINE = "@COLOR/CORE_ULTRAMARINE"
ULTRAMARINE_TINT = "#B0C8F9"
ULTRAMARINE_SHADE = "#1851B4"
GROUP = "@COLOR/CORE_ULTRAMARINE"
GROUP_TINT = "#B0C8F9"
GROUP_SHADE = "#1851B4"

COLORMAP = {
    "red": CRIMSON,
    "deep_orange": CRIMSON,
    "orange": VERMILLION,
    "amber": VERMILLION,
    "brown": BURLAP,
    "yellow": BURLAP,
    "pink": PLUMB,
    "purple": VIOLET,
    "deep_purple": VIOLET,
    "indigo": INDIGO,
    "blue": BLUE,
    "light_blue": BLUE,
    "cyan": TEAL,
    "teal": TEAL,
    "green": FOREST,
    "light_green": WINTERGREEN,
    "lime": WINTERGREEN,
    "blue_grey": TAUPE,
    "grey": STEEL,
    "ultramarine": ULTRAMARINE,
    "group_color": GROUP,
}


def get_random_color():
    return random.choice(list(COLORMAP.keys()))



# -*- coding: utf-8 -*-

"""Code for writing out the HTML

Author: Gertjan van den Burg

"""

import os
import datetime as dt

from emoji import emoji_lis as emoji_list
from jinja2 import Environment, PackageLoader, select_autoescape

from .models import MMSMessageRecord
from .types import (
    get_named_message_type,
    is_inbox_type,
    is_incoming_call,
    is_joined_type,
    is_missed_call,
    is_outgoing_call,
)
from .html_colors import COLORMAP


def is_all_emoji(body):
    """ Check if a message is non-empty and only contains emoji """
    body = body.replace(" ", "").replace("\ufe0f", "")
    return len(emoji_list(body)) == len(body) and len(body) > 0


def format_emoji(body, is_quote=False):
    """ Wrap emoji in <span> so we can style it easily """
    emoji_pos = emoji_list(body)
    new_body = ""
    emoji_lookup = {p["location"]: p["emoji"] for p in emoji_pos}
    for i, c in enumerate(body):
        if i in emoji_lookup:
            new_body += "<span class='msg-emoji'>%s</span>" % emoji_lookup[i]
        else:
            new_body += c
    return new_body


def dump_thread(thread, output_dir):
    """Write a Thread instance to a HTML page in the output directory """

    # Combine and sort the messages
    messages = thread.mms + thread.sms
    messages.sort(key=lambda mr: mr.dateSent)

    # Find the template
    env = Environment(
        loader=PackageLoader("signal2html", "templates"),
        autoescape=select_autoescape(["html", "xml"]),
    )
    template = env.get_template("thread.html")

    is_group = thread.recipient.isgroup

    # Create the message color CSS (depends on individuals)
    group_color_css = ""
    msg_css = ".msg-sender-%i { background: %s; }\n"
    if is_group:
        group_recipients = set(m.addressRecipient for m in messages)
        sender_idx = {r: k for k, r in enumerate(group_recipients)}
        colors_used = []
        group_colors = set(ar.color for ar in sender_idx)
        for ar, idx in sender_idx.items():
            if ar.recipientId._id.startswith("__textsecure_group__"):
                continue
            # ensure colors are unique, even if they're not in Signal
            ar_color = ar.color
            if ar_color in colors_used:
                color = next(
                    (c for c in COLORMAP if not c in group_colors),
                    None,
                )
                ar_color = ar.color if color is None else color
            group_color_css += msg_css % (idx, COLORMAP[ar_color])
            colors_used.append(ar.color)
    else:
        firstInbox = next(
            (m for m in messages if is_inbox_type(m._type)), None
        )
        clr = firstInbox.addressRecipient.color if firstInbox else "teal"
        clr = "teal" if clr is None else clr
        group_color_css += msg_css % (0, COLORMAP[clr])

    # Create a simplified dict for each message
    prev_date = None
    simple_messages = []
    for msg in messages:
        if is_joined_type(msg._type):
            continue

        # Add a "date change" message when to mark the date
        date_sent = dt.datetime.fromtimestamp(msg.dateSent // 1000)
        date_sent = date_sent.replace(microsecond=(msg.dateSent % 1000) * 1000)
        if prev_date is None or date_sent.date() != prev_date:
            prev_date = date_sent.date()
            out = {
                "date_msg": True,
                "body": date_sent.strftime("%a, %b %d, %Y"),
            }
            simple_messages.append(out)

        # Handle calls
        is_call = False
        if is_incoming_call(msg._type):
            is_call = True
            msg.body = f"{thread.name} called you"
        elif is_outgoing_call(msg._type):
            is_call = True
            msg.body = "You called"
        elif is_missed_call(msg._type):
            is_call = True
            msg.body = "Missed call"

        # Deal with quoted messages
        quote = {}
        if isinstance(msg, MMSMessageRecord) and msg.quote:
            quote_author_id = msg.quote.author.recipientId._id
            quote_author_name = msg.quote.author.name
            if quote_author_id == quote_author_name:
                name = "You"
            else:
                name = quote_author_name
            quote = {
                "name": name,
                "body": format_emoji(msg.quote.text),
                "attachments": [],
            }

        # Clean up message body
        body = "" if msg.body is None else msg.body
        if isinstance(msg, MMSMessageRecord):
            all_emoji = not msg.quote and is_all_emoji(body)
        else:
            all_emoji = is_all_emoji(body)
        body = format_emoji(body)

        # Create message dictionary
        aR = msg.addressRecipient
        out = {
            "isAllEmoji": all_emoji,
            "isGroup": is_group,
            "isCall": is_call,
            "type": get_named_message_type(msg._type),
            "body": body,
            "date": date_sent,
            "attachments": [],
            "id": msg._id,
            "name": aR.name[0],
            "sender_idx": sender_idx[aR] if is_group else "0",
            "quote": quote,
        }

        # Add attachments
        if isinstance(msg, MMSMessageRecord):
            for a in msg.attachments:
                if a.quote:
                    out["quote"]["attachments"].append(a)
                else:
                    out["attachments"].append(a)

        simple_messages.append(out)

    if not simple_messages:
        return

    html = template.render(
        thread_name=thread.name,
        messages=simple_messages,
        group_color_css=group_color_css,
    )

    os.makedirs(output_dir, exist_ok=True)

    filename = os.path.join(
        output_dir, thread.name.replace(" ", "_") + ".html"
    )
    with open(filename, "w", encoding="utf-8") as fp:
        fp.write(html)
